#!/usr/bin/python3

# axi-query-simple - apt-cache search replacement using apt-xapian-index
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options]",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import os
import warnings

import xapian

# This tells python-apt that we've seen the warning about the API not being
# stable yet, and we don't want to see every time we run the program
warnings.filterwarnings("ignore","apt API not stable yet")
import apt

warnings.resetwarnings()

# Setup configuration
XAPIANDBPATH = os.environ.get("AXI_DB_PATH", "/var/lib/apt-xapian-index")
XAPIANDB = XAPIANDBPATH + "/index"

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Stemmer function to generate stemmed search keywords
stemmer = xapian.Stem("english")

# Build the terms that will go in the query
terms = []
for word in args:
    if word.islower() and word.find("::") != -1:
        # If it's lowercase and it contains '::', then we consider it a Debtags
        # tag.  A better way could be to look up arguments in
        # /var/lib/debtags/vocabulary
        #
        # According to /var/lib/apt-xapian-index/README, Debtags tags are
        # indexed with the 'XT' prefix.
        terms.append("XT"+word)
    else:
        # If it is not a Debtags tag, then we consider it a normal keyword.
        word = word.lower()
        terms.append(word)
        # If the word has a stemmed version, add it to the query.
        # /var/lib/apt-xapian-index/README tells us that stemmed terms have a
        # 'Z' prefix.
        stem = stemmer(word)
        if stem != word:
            terms.append("Z"+stem)

# OR the terms together into a Xapian query.
#
# One may ask, why OR and not AND?  The reason is that, contrarily to
# apt-cache, Xapian scores results according to how well they matched.
#
# Matches that math all the terms will score higher than the others, so if we
# build an OR query what we really have is an AND query that gracefully
# degenerates to closer matches when they run out of perfect results.
#
# This allows stemmed searches to work nicely: if you look for 'editing', then
# the query will be 'editing OR Zedit'.  Packages with the word 'editing' will
# match both and score higher, and packages with the word 'edited' will still
# match 'Zedit' and be included in the results.
query = xapian.Query(xapian.Query.OP_OR, terms)

# Perform the query
enquire = xapian.Enquire(db)
enquire.set_query(query)

# Display the top 20 results, sorted by how well they match
cache = apt.Cache()
matches = enquire.get_mset(0, 20)
print("%i results found." % matches.get_matches_estimated())
print("Results 1-%i:" % matches.size())
for m in matches:
    # /var/lib/apt-xapian-index/README tells us that the Xapian document data
    # is the package name.
    name = m.document.get_data()

    # Get the package record out of the Apt cache, so we can retrieve the short
    # description
    pkg = cache[name]

    if pkg.candidate:
        # Print the match, together with the short description
        print("%i%% %s - %s" % (m.percent, name, pkg.candidate.summary))

sys.exit(0)



import os
import os.path

import xapian

AXI_ALIASES = os.environ.get("AXI_ALIASES", "/etc/apt-xapian-index/aliases/:/usr/share/apt-xapian-index/aliases/")

def read_db(progress=None):
    aliases = []
    maxts = 0
    files = []
    for d in AXI_ALIASES.split(":"):
        if not os.path.isdir(d): continue
        for f in os.listdir(d):
            if f[0] == '.': continue
            fname = os.path.join(d, f)
            ts = os.path.getmtime(fname)
            if ts > maxts:
                maxts = ts
            if progress: progress.verbose("Reading aliases from %s..." % fname)
            info = dict(path=fname)
            for idx, line in enumerate(open(fname)):
                line = line.strip()
                if idx == 0 and line[0] == '#':
                    # Take a comment at start of file as file description
                    info["desc"] = line[1:].strip()
                    continue
                # Skip comments and empty lines
                if not line or line[0] == '#': continue
                line = line.split()
                aliases.append(line)
            info.setdefault("desc", "synonyms for well-known terms")
            files.append(info)
    return maxts, aliases, files

class Aliases:
    def __init__(self, maxts, db, files):
        self.maxts = maxts
        self.db = db
        self.files = files

    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:

         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        return dict(timestamp=self.maxts, sources=self.files)

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """
        pass

    def send_extra_info(self, db=None, **kw):
        """
        Receive extra parameters from the indexer.

        This may be called more than once, but after init().

        We are using this to get the database instance
        """
        if db is not None:
            for row in self.db:
                for a in row[1:]:
                    db.add_synonym(row[0], a)

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        return dict(
            name = "Package aliases",
            shortDesc = "aliases for well known programs",
            fullDoc = """
            The Aliases data source does not change documents in the index, but
            adds synonims to the database. Synonims allow to obtain good
            results while looking for well-know software names, even if such
            software does not exist in Debian.
            """
        )

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        pass

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        pass

def init(progress=None, **kw):
    """
    Create and return the plugin object.
    """
    maxts, db, files = read_db(progress)
    if not db: return None
    return Aliases(maxts, db, files)



class Template:
    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:
         
         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.
         sources (optional)
           ad array of dicts { path: pathname, desc: description }, one for
           every data file accessed by this plugin. A directory can be provided
           as path, meaning "it accesses all sorts of files inside this
           directory": for example the APT index, or the app-install-data
           files. Use [] to mean "no sources".

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        return dict(timestamp = 0, values = [], sources=[])

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """
        pass

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        #return dict(
        #    name = "template",
        #    shortDesc = "Example plugin that does nothing",
        #    fullDoc = """
        #    Here goes full documentation in ReST format
        #    """
        #)

        # Return None if you don't want this file to appear in the
        # documentation.  Probably, only the Template plugin should do that.
        return None

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        pass

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        pass

    def finished(self):
        """
        Called when the indexing is finihsed
        """


def init(**kw):
    """
    Create and return the plugin object.

    Keyword arguments can hold various kind of information passed by the
    indexer. Currently documented is:

     - langs: sequence of languages for which we are building the index

    Return None here to disable this plugin.
    """
    return Template()



# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.

import os
import re
import sys
import warnings

import xapian

# Setup configuration
# This tells python-apt that we've seen the warning about the API not being
# stable yet, and we don't want to see every time we run the program
warnings.filterwarnings("ignore","apt API not stable yet")
import apt

warnings.resetwarnings()


# Setup configuration
XAPIANDBPATH = os.environ.get("AXI_DB_PATH", "/var/lib/apt-xapian-index")
XAPIANDB = XAPIANDBPATH + "/index"
XAPIANDBVALUES = XAPIANDBPATH + "/values"

# This is our little database of simple Debtags filters we provide: the name
# entered by the user in "--type" maps to a piece of Xapian query
filterdb = dict(
    # We can do simple AND queries...
    game = xapian.Query(xapian.Query.OP_AND, ('XTuse::gameplaying', 'XTrole::program')),
    # Or we can do complicate binary expressions...
    gui = xapian.Query(xapian.Query.OP_AND, xapian.Query('XTrole::program'),
                xapian.Query(xapian.Query.OP_OR, 'XTinterface::x11', 'XTinterface::3d')),
    cmdline = xapian.Query(xapian.Query.OP_AND, 'XTrole::program', 'XTinterface::commandline'),
    editor = xapian.Query(xapian.Query.OP_AND, 'XTrole::program', 'XTuse::editing')
    # Feel free to invent more
)

def termsForSimpleQuery(keywords):
    """
    Given a list of user-supplied keywords, build the list of terms that will
    go in a simple Xapian query.

    If a term is lowercase and contains '::', then it's considered to be a
    Debtags tag.
    """
    stemmer = xapian.Stem("english")
    terms = []
    for word in keywords:
        if word.islower() and word.find("::") != -1:
            # FIXME: A better way could be to look up arguments in
            # /var/lib/debtags/vocabulary
            #
            # According to /var/lib/apt-xapian-index/README, Debtags tags are
            # indexed with the 'XT' prefix.
            terms.append("XT"+word)
        else:
            # If it is not a Debtags tag, then we consider it a normal keyword.
            word = word.lower()
            terms.append(word)
            # If the word has a stemmed version, add it to the query.
            # /var/lib/apt-xapian-index/README tells us that stemmed terms have a
            # 'Z' prefix.
            stem = stemmer(word)
            if stem != word:
                terms.append("Z"+stem)
    return terms

def addSimpleFilterToQuery(query, filtername):
    """
    If filtername is not None, lookup the simple filter database for the name
    and add its filter to the query.  Returns the enhanced query.
    """
    # See if the user wants to use one of the result filters
    if filtername:
        if filtername in filterdb:
            # If a filter was requested, AND it with the query
            return xapian.Query(xapian.Query.OP_AND, filterdb[filtername], query)
        else:
            raise RuntimeError("Invalid filter type.  Try one of " + ", ".join(sorted(filterdb)))
    else:
        return query

def show_mset(mset):
    """
    Show a Xapian result mset as a list of packages and their short descriptions
    """
    # Display the top 20 results, sorted by how well they match
    cache = apt.Cache()
    print("%i results found." % mset.get_matches_estimated())
    print("Results 1-%i:" % mset.size())
    for m in mset:
        # /var/lib/apt-xapian-index/README tells us that the Xapian document data
        # is the package name.
        name = m.document.get_data()

        # Get the package record out of the Apt cache, so we can retrieve the short
        # description
        pkg = cache[name]

        # Print the match, together with the short description
        if pkg.candidate:
            print("%i%% %s - %s" % (m.percent, name, pkg.candidate.summary))

def readValueDB(pathname):
    """
    Read the "/etc/services"-style database of value indices
    """
    try:
        rmcomments = re.compile("\s*(#.*)?$")
        splitter = re.compile("\s+")
        values = {}
        for idx, line in enumerate(open(pathname)):
            # Remove comments and trailing spaces
            line = rmcomments.sub("", line)
            # Skip empty lines
            if len(line) == 0: continue
            # Split the line
            fields = splitter.split(line)
            if len(fields) < 2:
                print("Ignoring line %s:%d: only 1 value found when I need at least the value name and number" % (pathname, idx+1), file=sys.stderr)
                continue
            # Parse the number
            try:
                number = int(fields[1])
            except ValueError:
                print("Ignoring line %s:%d: the second column (\"%s\") must be a number" % (pathname, idx+1, fields[1]), file=sys.stderr)
                continue
            values[fields[0]] = number
            for alias in fields[2:]:
                values[alias] = number
    except OSError as e:
        # If we can't read the database, fallback to defaults
        print("Cannot read %s: %s.  Using a minimal default configuration" % (pathname, e), file=sys.stderr)
        values = dict(
            installedsize = 1,
            packagesize = 2
        )
    return values



#!/usr/bin/python3

#
# axi-query - Example program to query the apt-xapian-index
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
from optparse import OptionParser

from aptxapianindex import (
    XAPIANDB,
    XAPIANDBVALUES,
    readValueDB,
)

VERSION="0.1"

class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options]",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")
parser.add_option("-s", "--sort", help="sort by the given value, as listed in %s" % XAPIANDBVALUES)

(options, args) = parser.parse_args()


import warnings

import xapian

# Yes, apt, thanks, I know, the api isn't stable, thank you so very much
#warnings.simplefilter('ignore', FutureWarning)
warnings.filterwarnings("ignore","apt API not stable yet")
import apt

warnings.resetwarnings()

# Access the Xapian index
db = xapian.Database(XAPIANDB)

# Build the query
stemmer = xapian.Stem("english")
terms = []
for word in args:
    if word.islower() and word.find("::") != -1:
        # If it's lowercase and contains, :: it's a tag
        # TODO: lookup in debtags' vocabulary instead
        terms.append("XT"+word)
    else:
        # Else we make a term
        word = word.lower()
        terms.append(word)
        stem = stemmer(word)
        # If it has stemming, add that to the query, too
        if stem != word:
            terms.append("Z"+stem)
query = xapian.Query(xapian.Query.OP_OR, terms)

# Perform the query
enquire = xapian.Enquire(db)
enquire.set_query(query)
if options.sort:
    values = readValueDB(XAPIANDBVALUES)

    # If we don't sort by relevance, we need to specify a cutoff in order to
    # remove poor results from the output
    #
    # Note: ept-cache implements an adaptive cutoff as follows:
    # 1. Retrieve only one result, with default sorting.  Read its relevance as
    #    the maximum relevance.
    # 2. Set the cutoff as some percentage of the maximum relevance
    # 3. Set sort by the wanted value
    # 4. Perform the query
    enquire.set_cutoff(60)

    # Sort by the requested value
    enquire.set_sort_by_value(values[options.sort])

# Display the results.
cache = apt.Cache()
matches = enquire.get_mset(0, 20)
print("%i results found." % matches.get_matches_estimated())
print("Results 1-%i:" % matches.size())
for m in matches:
    name = m.document.get_data()
    pkg = cache[name]
    if pkg.candidate:
        print("%i%% %s - %s" % (m.percent, name, pkg.candidate.summary))

sys.exit(0)



# -*- coding: utf-8 -*-
from __future__ import annotations

import os
import shutil
import subprocess
import unittest
from typing import Iterator

import tools

import axi
import axi.indexer


def smallcache(pkglist=["apt", "libept-dev", "gedit"]):
    class sc:
        def __init__(self, cache) -> None:
            self._pkgs = pkglist
            self._cache = cache

        def has_key(self, name: str) -> bool:
            return name in self._pkgs

        def __len__(self) -> int:
            return len(self._pkgs)

        def __iter__(self) -> Iterator[str]:
            for p in self._pkgs:
                yield self._cache[p]

        def __getitem__(self, name: str):
            if name not in self._pkgs:
                raise KeyError("`%s' not in wrapped cache" % name)
            return self._cache[name]

    return sc

class TestIndexer(tools.AxiTestBase):
    indexer: axi.indexer.Indexer | None

    def setUp(self) -> None:
        # Remove the text index if it exists
        if os.path.exists(axi.XAPIANDBPATH): shutil.rmtree(axi.XAPIANDBPATH)
        # Prepare a quiet indexer
        progress = axi.indexer.SilentProgress()
        self.indexer = axi.indexer.Indexer(progress, True)

    def tearDown(self) -> None:
        # Explicitly set indexer to none, otherwise in the next setUp we rmtree
        # testdb before the indexer had a chance to delete its lock
        self.indexer = None

    def testAptRebuild(self) -> None:
        assert self.indexer
        self.indexer._test_wrap_apt_cache(smallcache())

        # No other indexers are running, ensure lock succeeds
        self.assert_(self.indexer.lock())

        # No index exists, so the indexer should decide it needs to run
        self.assert_(self.indexer.setupIndexing())

        # Perform a rebuild
        self.indexer.rebuild()

        # Close the indexer
        self.indexer = None

        # Ensure that we have an index
        self.assertCleanIndex()

    def testDeb822Rebuild(self) -> None:
        assert self.indexer
        pkgfile = os.path.join(axi.XAPIANDBPATH, "packages")
        subprocess.check_call("apt-cache show apt libept-dev gedit > " + pkgfile, shell=True)

        # No other indexers are running, ensure lock succeeds
        self.assert_(self.indexer.lock())

        # No index exists, so the indexer should decide it needs to run
        self.assert_(self.indexer.setupIndexing())

        # Perform a rebuild
        self.indexer.rebuild([pkgfile])

        # Close the indexer
        self.indexer = None

        # Ensure that we have an index
        self.assertCleanIndex()

    def testIncrementalRebuild(self) -> None:
        assert self.indexer
        pre_indexer: axi.indexer.Indexer | None
        # Perform the initial indexing
        progress = axi.indexer.SilentProgress()
        pre_indexer = axi.indexer.Indexer(progress, True)
        pre_indexer._test_wrap_apt_cache(smallcache(["apt", "libept-dev", "gedit"]))
        self.assert_(pre_indexer.lock())
        self.assert_(pre_indexer.setupIndexing())
        pre_indexer.rebuild()
        pre_indexer = None
        curidx = open(axi.XAPIANINDEX).read()

        # Ensure that we have an index
        self.assertCleanIndex()

        # Prepare an incremental update
        self.indexer._test_wrap_apt_cache(smallcache(["apt", "coreutils", "gedit"]))

        # No other indexers are running, ensure lock succeeds
        self.assert_(self.indexer.lock())

        # An index exists the plugin modification timestamps are the same, so
        # we need to force the indexer to run
        self.assert_(not self.indexer.setupIndexing())
        self.assert_(self.indexer.setupIndexing(force=True))

        # Perform a rebuild
        self.indexer.incrementalUpdate()

        # Close the indexer
        self.indexer = None

        # Ensure that we have an index
        self.assertCleanIndex()

        # Ensure that we did not create a new index
        self.assertEqual(open(axi.XAPIANINDEX).read(), curidx)

    def testIncrementalRebuildFromEmpty(self) -> None:
        assert self.indexer
        # Prepare an incremental update
        self.indexer._test_wrap_apt_cache(smallcache())

        # No other indexers are running, ensure lock succeeds
        self.assert_(self.indexer.lock())

        # No index exists, so the indexer should decide it needs to run
        self.assert_(self.indexer.setupIndexing())

        # Perform an incremental rebuild, which should fall back on a normal
        # one
        self.indexer.incrementalUpdate()

        # Close the indexer
        self.indexer = None

        # Ensure that we have an index
        self.assertCleanIndex()

#    def test_url(self):
#        """ Environ: URL building """
#        request.bind({'HTTP_HOST':'example.com'}, None)
#        self.assertEqual('http://example.com/', request.url)
#        request.bind({'SERVER_NAME':'example.com'}, None)
#        self.assertEqual('http://example.com/', request.url)
#        request.bind({'SERVER_NAME':'example.com', 'SERVER_PORT':'81'}, None)
#        self.assertEqual('http://example.com:81/', request.url)
#        request.bind({'wsgi.url_scheme':'https', 'SERVER_NAME':'example.com'}, None)
#        self.assertEqual('https://example.com:80/', request.url)
#        request.bind({'HTTP_HOST':'example.com', 'PATH_INFO':'/path', 'QUERY_STRING':'1=b&c=d', 'SCRIPT_NAME':'/sp'}, None)
#        self.assertEqual('http://example.com/sp/path?1=b&c=d', request.url)
#
#    def test_dict_access(self):
#        """ Environ: request objects are environment dicts """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        request.bind(e, None)
#        for k, v in e.iteritems():
#            self.assertTrue(k in request)
#            self.assertTrue(request[k] == v)
#
#    def test_header_access(self):
#        """ Environ: Request objects decode headers """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['HTTP_SOME_HEADER'] = 'some value'
#        request.bind(e, None)
#        request['HTTP_SOME_OTHER_HEADER'] = 'some other value'
#        self.assertTrue('Some-Header' in request.header)
#        self.assertTrue(request.header['Some-Header'] == 'some value')
#        self.assertTrue(request.header['Some-Other-Header'] == 'some other value')
#
#
#    def test_cookie(self):
#        """ Environ: COOKIES """
#        t = dict()
#        t['a=a'] = {'a': 'a'}
#        t['a=a; b=b'] = {'a': 'a', 'b':'b'}
#        t['a=a; a=b'] = {'a': 'b'}
#        for k, v in t.iteritems():
#            request.bind({'HTTP_COOKIE': k}, None)
#            self.assertEqual(v, request.COOKIES)
#
#    def test_get(self):
#        """ Environ: GET data """
#        e = {}
#        e['QUERY_STRING'] = 'a=a&a=1&b=b&c=c'
#        request.bind(e, None)
#        self.assertTrue('a' in request.GET)
#        self.assertTrue('b' in request.GET)
#        self.assertEqual(['a','1'], request.GET.getall('a'))
#        self.assertEqual(['b'], request.GET.getall('b'))
#        self.assertEqual('1', request.GET['a'])
#        self.assertEqual('b', request.GET['b'])
#
#    def test_post(self):
#        """ Environ: POST data """
#        sq = u'a=a&a=1&b=b&c=c'.encode('utf8')
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['wsgi.input'].write(sq)
#        e['wsgi.input'].seek(0)
#        e['CONTENT_LENGTH'] = str(len(sq))
#        e['REQUEST_METHOD'] = "POST"
#        request.bind(e, None)
#        self.assertTrue('a' in request.POST)
#        self.assertTrue('b' in request.POST)
#        self.assertEqual(['a','1'], request.POST.getall('a'))
#        self.assertEqual(['b'], request.POST.getall('b'))
#        self.assertEqual('1', request.POST['a'])
#        self.assertEqual('b', request.POST['b'])
#
#    def test_params(self):
#        """ Environ: GET and POST are combined in request.param """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['wsgi.input'].write(tob('b=b&c=p'))
#        e['wsgi.input'].seek(0)
#        e['CONTENT_LENGTH'] = '7'
#        e['QUERY_STRING'] = 'a=a&c=g'
#        e['REQUEST_METHOD'] = "POST"
#        request.bind(e, None)
#        self.assertEqual(['a','b','c'], sorted(request.params.keys()))
#        self.assertEqual('p', request.params['c'])
#
#    def test_getpostleak(self):
#        """ Environ: GET and POST should not leak into each other """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['wsgi.input'].write(u'b=b'.encode('utf8'))
#        e['wsgi.input'].seek(0)
#        e['CONTENT_LENGTH'] = '3'
#        e['QUERY_STRING'] = 'a=a'
#        e['REQUEST_METHOD'] = "POST"
#        request.bind(e, None)
#        self.assertEqual(['a'], request.GET.keys())
#        self.assertEqual(['b'], request.POST.keys())
#
#    def test_body(self):
#        """ Environ: Request.body should behave like a file object factory """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['wsgi.input'].write(u'abc'.encode('utf8'))
#        e['wsgi.input'].seek(0)
#        e['CONTENT_LENGTH'] = str(3)
#        request.bind(e, None)
#        self.assertEqual(u'abc'.encode('utf8'), request.body.read())
#        self.assertEqual(u'abc'.encode('utf8'), request.body.read(3))
#        self.assertEqual(u'abc'.encode('utf8'), request.body.readline())
#        self.assertEqual(u'abc'.encode('utf8'), request.body.readline(3))
#
#    def test_bigbody(self):
#        """ Environ: Request.body should handle big uploads using files """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['wsgi.input'].write((u'x'*1024*1000).encode('utf8'))
#        e['wsgi.input'].seek(0)
#        e['CONTENT_LENGTH'] = str(1024*1000)
#        request.bind(e, None)
#        self.assertTrue(hasattr(request.body, 'fileno'))
#        self.assertEqual(1024*1000, len(request.body.read()))
#        self.assertEqual(1024, len(request.body.read(1024)))
#        self.assertEqual(1024*1000, len(request.body.readline()))
#        self.assertEqual(1024, len(request.body.readline(1024)))
#
#    def test_tobigbody(self):
#        """ Environ: Request.body should truncate to Content-Length bytes """
#        e = {}
#        wsgiref.util.setup_testing_defaults(e)
#        e['wsgi.input'].write((u'x'*1024).encode('utf8'))
#        e['wsgi.input'].seek(0)
#        e['CONTENT_LENGTH'] = '42'
#        request.bind(e, None)
#        self.assertEqual(42, len(request.body.read()))
#        self.assertEqual(42, len(request.body.read(1024)))
#        self.assertEqual(42, len(request.body.readline()))
#        self.assertEqual(42, len(request.body.readline(1024)))
#
#class TestMultipart(unittest.TestCase):
#    def test_multipart(self):
#        """ Environ: POST (multipart files and multible values per key) """
#        fields = [('field1','value1'), ('field2','value2'), ('field2','value3')]
#        files = [('file1','filename1.txt','content1'), ('file2','filename2.py',u'äöü')]
#        e = tools.multipart_environ(fields=fields, files=files)
#        request.bind(e, None)
#        # File content
#        self.assertTrue('file1' in request.POST)
#        self.assertEqual('content1', request.POST['file1'].file.read())
#        # File name and meta data
#        self.assertTrue('file2' in request.POST)
#        self.assertEqual('filename2.py', request.POST['file2'].filename)
#        # UTF-8 files
#        x = request.POST['file2'].file.read()
#        if sys.version_info >= (3,0,0):
#            x = x.encode('ISO-8859-1')
#        self.assertEqual(u'äöü'.encode('utf8'), x)
#        # No file
#        self.assertTrue('file3' not in request.POST)
#        # Field (single)
#        self.assertEqual('value1', request.POST['field1'])
#        # Field (multi)
#        self.assertEqual(2, len(request.POST.getall('field2')))
#        self.assertEqual(['value2', 'value3'], request.POST.getall('field2'))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

# axi-query-similar - Show packages similar to a given one
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options] package(s)",
                version="%prog "+ VERSION,
                description="Find the packages similar to the given ones")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import xapian
from aptxapianindex import (
    XAPIANDB,
    show_mset,
)

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

def docForPackage(pkgname):
    "Get the document corresponding to the package with the given name"
    # Query the term with the package name
    query = xapian.Query("XP"+pkgname)
    enquire = xapian.Enquire(db)
    enquire.set_query(query)
    # Get the top result only
    matches = enquire.get_mset(0, 1)
    if matches.size() == 0:
        return None
    else:
        m = matches[0]
        return m.document

# Build a term list with all the terms in the given packages
terms = []
for pkgname in args:
    # Get the document corresponding to the package name
    doc = docForPackage(pkgname)
    if not doc: continue
    # Retrieve all the terms in the document
    for t in doc.termlist():
        t = t.decode('utf-8')
        if len(t.term) < 2 or t.term[:2] != 'XP':
            terms.append(t.term)

# Build the big OR query
query = xapian.Query(xapian.Query.OP_AND_NOT,
            # Terms we want
            xapian.Query(xapian.Query.OP_OR, terms),
            # AND NOT the input packages
            xapian.Query(xapian.Query.OP_OR, ["XP"+name for name in args]))

# Perform the query
enquire = xapian.Enquire(db)
enquire.set_query(query)

# Retrieve the top 20 results
matches = enquire.get_mset(0, 20)

# Display the results
show_mset(matches)

sys.exit(0)



try:
    import apt
    HAS_APT=True
except ImportError:
    HAS_APT=False
import codecs
import os
import re
from urllib.parse import unquote

import xapian

from debian import deb822

APTLISTDIR="/var/lib/apt/lists"

def translationFiles(langs=None):
    # Look for files like: ftp.uk.debian.org_debian_dists_sid_main_i18n_Translation-it
    # And extract the language code at the end
    tfile = re.compile(r"_i18n_Translation-([^-]+)$")
    for f in os.listdir(APTLISTDIR):
        mo = tfile.search(f)
        if not mo: continue
        if langs and not mo.group(1) in langs: continue
        yield unquote(mo.group(1)), os.path.join(APTLISTDIR, f)

class Indexer:
    def __init__(self, lang, file):
        self.lang = lang
        self.xlang = lang.split("_")[0]
        self.indexer = xapian.TermGenerator()
        # Get a stemmer for this language, if available
        try:
            self.stemmer = xapian.Stem(self.xlang)
            self.indexer.set_stemmer(self.stemmer)
        except xapian.InvalidArgumentError:
            pass

        # Read the translated descriptions
        self.descs = dict()
        desckey = "Description-"+self.lang
        with codecs.open(file, 'r', encoding='utf-8') as fp:
            for pkg in deb822.Deb822.iter_paragraphs(fp):
                # I need this if because in some translation files, some
                # packages have a different Description header.  For example,
                # in the -de translations, I once found a
                # Description-de.noguide: header instead of Description-de:
                if desckey in pkg:
                    self.descs[pkg["Package"]] = pkg[desckey]

    def index(self, document):
        name = document.get_data()
        self.indexer.set_document(document)
        self.indexer.index_text_without_positions(self.descs.get(name, ""))

class TranslatedDescriptions:
    def __init__(self, langs):
        self.langs = langs

    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:
         
         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        if not HAS_APT: return dict(timestamp = 0)
        filelist = translationFiles(self.langs)
        maxts = max([0] + [os.path.getmtime(f) for l, f in filelist])
        return dict(
                timestamp=maxts,
                sources=[dict(path=f, desc="%s translation" % l) for l, f in filelist],
                prefixes=[
                    dict(idx="Z", qp=None, type=None,
                         desc="Stemmed forms of keywords",
                         ldesc="This contains the stemmed forms of keywords as generated by"
                               " TermGenerator and matched by QueryParser"),
                ],
        )

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """

        self.indexers = []
        for lang, file in translationFiles(self.langs):
            progress.begin("Reading %s translations from %s" % (lang, file))
            self.indexers.append(Indexer(lang, file))
            progress.end()

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        return dict(
            name = "Translated package descriptions",
            shortDesc = "terms extracted from the translated package descriptions using Xapian's TermGenerator",
            fullDoc = """
            The TranslatedDescriptions data source reads translated description
            files from %s, then uses Xapian's TermGenerator to tokenise and
            index their content.

            Currently this creates normal terms as well as stemmed terms
            prefixed with ``Z``.
            """ % APTLISTDIR
        )

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        for i in self.indexers:
            i.index(document)

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        for i in self.indexers:
            i.index(document)

def init(langs=None, **kw):
    """
    Create and return the plugin object.
    """
    if not HAS_APT: return None
    if not langs: return None
    files = [f for l, f in translationFiles(langs)]
    if len(files) == 0:
        return None
    return TranslatedDescriptions(langs=langs)



# -*- coding: utf-8 -*-
#
# axi/indexer.py - apt-xapian-index indexer
#
# Copyright (C) 2007--2010  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from __future__ import annotations

import errno
import fcntl
import itertools
import os
import pickle
import re
import shutil
import socket
import sys
import textwrap
import time
from importlib.machinery import SourceFileLoader
from operator import itemgetter
from types import TracebackType
from typing import TYPE_CHECKING, Any, Iterable, Iterator
from urllib.parse import unquote

import xapian

import axi

if TYPE_CHECKING:
    import apt
    import apt_pkg

APTLISTDIR="/var/lib/apt/lists"

class Addon:
    """
    Indexer plugin wrapper
    """
    def __init__(
        self,
        fname: str,
        progress=None,
        **kw
    ) -> None:
        self.filename = os.path.basename(fname)
        self.name = os.path.splitext(self.filename)[0]
        oldpath = sys.path
        try:
            sys.path.append(os.path.dirname(fname))
            self.module = SourceFileLoader("axi.plugin_" + self.name, fname).load_module()
        finally:
            sys.path = oldpath
        try:
            self.obj = self.module.init(**kw)
        except TypeError:
            self.obj = self.module.init()
        if self.obj:
            try:
                try:
                    self.info = self.obj.info(**kw)
                except TypeError:
                    self.info = self.obj.info()
            except Exception as e:
                if progress:
                    progress.warning("Plugin %s initialisation failed: %s" % (fname, str(e)))
                self.obj = None

    def finished(self) -> None:
        if hasattr(self.obj, "finished"):
            self.obj.finished()

    def send_extra_info(self, **kw: Any) -> None:
        func = getattr(self.obj, "send_extra_info", None)
        if func is not None:
            func(**kw)

class Plugins(list):
    def __init__(self, **kw) -> None:
        """
        Read the plugins, in sorted order.

        Pass all the keyword args to the plugin init
        """
        if "langs" not in kw:
            kw["langs"] = self.scan_available_languages()
        progress = kw.get("progress", None)

        self.disabled: list[str] = []
        for fname in sorted(os.listdir(axi.PLUGINDIR)):
            # Skip non-pythons, hidden files and python sources starting with '_'
            if fname[0] in ['.', '_'] or not fname.endswith(".py"):
                continue
            fullname = os.path.join(axi.PLUGINDIR, fname)
            if not os.path.isfile(fullname):
                continue
            if progress:
                progress.verbose("Reading plugin %s." % fullname)
            addon = Addon(fullname, **kw)
            if addon.obj != None:
                self.append(addon)

    def scan_available_languages(self) -> set[str]:
        # Languages we index
        langs: set[str] = set()

        # Look for files like: ftp.uk.debian.org_debian_dists_sid_main_i18n_Translation-it
        # And extract the language code at the end
        tfile = re.compile(r"_i18n_Translation-([^-]+)$")
        for f in os.listdir(APTLISTDIR):
            mo = tfile.search(f)
            if not mo: continue
            langs.add(unquote(mo.group(1)))

        return langs


class Progress:
    """
    Normal progress report to stdout
    """
    def __init__(self) -> None:
        self.task: str | None = None
        self.halfway = False
        self.is_verbose = False

    def begin(self, task: str) -> None:
        self.task = task
        print("%s..." % self.task, end='')
        sys.stdout.flush()
        self.halfway = True

    def progress(self, percent: float) -> None:
        print("\r%s... %d%%" % (self.task, percent), end='')
        sys.stdout.flush()
        self.halfway = True

    def end(self) -> None:
        print("\r%s: done.  " % self.task)
        self.halfway = False

    def verbose(self, *args: str) -> None:
        if not self.is_verbose: return
        if self.halfway:
            print()
        print(" ".join(args))
        self.halfway = False

    def notice(self, *args: str) -> None:
        if self.halfway:
            print()
        print(" ".join(args), file=sys.stderr)
        self.halfway = False

    def warning(self, *args: str) -> None:
        if self.halfway:
            print()
        print(" ".join(args), file=sys.stderr)
        self.halfway = False

    def error(self, *args: str) -> None:
        if self.halfway:
            print()
        print(" ".join(args), file=sys.stderr)
        self.halfway = False


class BatchProgress:
    """
    Machine readable progress report
    """
    def __init__(self) -> None:
        self.task: str | None = None

    def begin(self, task: str) -> None:
        self.task = task
        print("begin: %s\n" % self.task, end='')
        sys.stdout.flush()

    def progress(self, percent: float) -> None:
        print("progress: %d/100\n" % percent, end='')
        sys.stdout.flush()

    def end(self) -> None:
        print("done: %s\n" % self.task)
        sys.stdout.flush()

    def verbose(self, *args: str) -> None:
        print("verbose: %s" % (" ".join(args)))
        sys.stdout.flush()

    def notice(self, *args: str) -> None:
        print("notice: %s" % (" ".join(args)))
        sys.stdout.flush()

    def warning(self, *args: str) -> None:
        print("warning: %s" % (" ".join(args)))
        sys.stdout.flush()

    def error(self, *args:str) -> None:
        print("error: %s" % (" ".join(args)))
        sys.stdout.flush()


class SilentProgress:
    """
    Quiet progress report
    """
    def begin(self, task: str) -> None:
        pass

    def progress(self, percent: float) -> None:
        pass

    def end(self) -> None:
        pass

    def verbose(self, *args: str) -> None:
        pass

    def notice(self, *args: str) -> None:
        pass
    def warning(self, *args: str) -> None:
        print(" ".join(args), file=sys.stderr)

    def error(self, *args: str) -> None:
        print(" ".join(args), file=sys.stderr)


class ClientProgress:
    """
    Client-side progress report, reporting progress from another running
    indexer
    """
    def __init__(self, progress) -> None:
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.settimeout(None)
        self.sock.connect(axi.XAPIANDBUPDATESOCK)
        self.progress = progress

    def loop(self) -> None:
        hasBegun = False
        while True:
            msg = self.sock.recv(4096)
            try:
                args = pickle.loads(msg)
            except EOFError:
                self.progress.error("The other update has stopped")
                return
            action = args[0]
            args = args[1:]
            if action == "begin":
                self.progress.begin(*args)
                hasBegun = True
            elif action == "progress":
                if not hasBegun:
                    self.progress.begin(args[0])
                    hasBegun = True
                self.progress.progress(*args[1:])
            elif action == "end":
                if not hasBegun:
                    self.progress.begin(args[0])
                    hasBegun = True
                self.progress.end(*args[1:])
            elif action == "verbose":
                self.progress.verbose(*args)
            elif action == "notice":
                self.progress.notice(*args)
            elif action == "error":
                self.progress.error(*args)
            elif action == "alldone":
                break
            else:
                self.progress.error("unknown action '%s' from other update-apt-xapian-index.  Arguments: '%s'" % (action, ", ".join(map(repr, args))))


class ServerSenderProgress:
    """
    Server endpoint for client-server progress report
    """
    def __init__(self, sock: socket.socket, task: str | None = None) -> None:
        self.sock = sock
        self.task: str | None = task

    def __del__(self) -> None:
        self._send(pickle.dumps(("alldone",)))

    def _send(self, text: bytes) -> None:
        try:
            self.sock.send(text)
        except:
            pass

    def begin(self, task: str) -> None:
        self.task = task
        self._send(pickle.dumps(("begin", self.task)))

    def progress(self, percent: float) -> None:
        self._send(pickle.dumps(("progress", self.task, percent)))

    def end(self) -> None:
        self._send(pickle.dumps(("end", self.task)))

    def verbose(self, *args: str) -> None:
        self._send(pickle.dumps(("verbose",) + args))

    def notice(self, *args: str) -> None:
        self._send(pickle.dumps(("notice",) + args))

    def warning(self, *args: str) -> None:
        self._send(pickle.dumps(("warning",) + args))

    def error(self, *args: str) -> None:
        self._send(pickle.dumps(("error",) + args))


class ServerProgress:
    """
    Send progress report to any progress object, as well as to client indexers
    """
    def __init__(self, mine) -> None:
        self.task: str | None = None
        self.proxied = [mine]
        self.sockfile = axi.XAPIANDBUPDATESOCK
        try:
            os.unlink(self.sockfile)
        except OSError:
            pass
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(axi.XAPIANDBUPDATESOCK)
        self.sock.setblocking(False)
        self.sock.listen(5)
        # Disallowing unwanted people to mess with the file is automatic, as
        # the socket has the ownership of the user we're using, and people
        # can't connect to it unless they can write to it

    def __del__(self) -> None:
        self.sock.close()
        os.unlink(self.sockfile)

    def _check(self) -> None:
        try:
            sock = self.sock.accept()[0]
            self.proxied.append(ServerSenderProgress(sock, self.task))
        except socket.error as e:
            if e.args[0] != errno.EAGAIN:
                raise
        pass

    def begin(self, task: str) -> None:
        self._check()
        self.task = task
        for x in self.proxied: x.begin(task)

    def progress(self, percent: float) -> None:
        self._check()
        for x in self.proxied: x.progress(percent)

    def end(self) -> None:
        self._check()
        for x in self.proxied: x.end()

    def verbose(self, *args: str) -> None:
        self._check()
        for x in self.proxied: x.verbose(*args)

    def notice(self, *args: str) -> None:
        self._check()
        for x in self.proxied: x.notice(*args)

    def warning(self, *args: str) -> None:
        self._check()
        for x in self.proxied: x.warning(*args)

    def error(self, *args: str) -> None:
        self._check()
        for x in self.proxied: x.error(*args)


class ExecutionTime:
    """
    Helper that can be used in with statements to have a simple
    measure of the timing of a particular block of code, e.g.
    with ExecutionTime("db flush"):
        db.flush()
    """
    def __init__(self, info: str = "") -> None:
        self.info = info

    def __enter__(self) -> None:
        self.now = time.time()

    def __exit__(
        self,
        type : type[BaseException] | None = None,
        value: BaseException | None = None,
        stack: TracebackType | None = None
    ) -> None:
        print("%s: %s" % (self.info, time.time() - self.now))


class Indexer:
    """
    The indexer
    """
    def __init__(self, progress, quietapt: bool = False) -> None:
        self.progress = progress
        self.quietapt = quietapt
        self.verbose = getattr(progress, "is_verbose", False)
        # Timestamp of the most recent data source
        self.ds_timestamp = 0
        # Apt cache instantiated on demand
        self.apt_cache: apt.Cache | None = None
        # Loaded plugins
        self.plugins: list[Addon] | None = None
        # OS file handle for the lock file
        self.lockfd: int | None = None

        # Ensure the database and cache directories exist
        self.ensure_dir_exists(axi.XAPIANDBPATH)
        self.ensure_dir_exists(axi.XAPIANCACHEPATH)

    def ensure_dir_exists(self, pathname: str) -> None:
        """
        Create a directory if missing, but do not complain if it already exists
        """
        try:
            # Try to create it anyway
            os.mkdir(pathname)
        except OSError as e:
            if e.errno != errno.EEXIST:
                # If we got an error besides path already existing, fail
                raise
            elif not os.path.isdir(pathname):
                # If that path already exists, but is not a directory, also fail
                raise

    def _test_wrap_apt_cache(self, wrapper) -> None:
        """
        Wrap the apt-cache in some proxy object.

        This is used to give tests some control over the apt cache results
        """
        if self.apt_cache is not None:
            raise RuntimeError("the cache has already been instantiated")
        # Instantiate the cache
        self.aptcache()
        # Wrap it
        self.apt_cache = wrapper(self.apt_cache)

    def aptcache(self) -> apt.Cache:
        if not self.apt_cache:
            #import warnings
            ## Yes, apt, thanks, I know, the api isn't stable, thank you so very much
            ##warnings.simplefilter('ignore', FutureWarning)
            #warnings.filterwarnings("ignore","apt API not stable yet")
            import apt
            import apt_pkg

            #warnings.resetwarnings()

            if self.quietapt:
                class AptSilentProgress(apt.progress.text.OpProgress) :
                    def __init__(self) -> None:
                        pass
                    def done(self) -> None:
                        pass
                    def update(self, percent: float | None = None) -> None:
                        pass
                aptprogress = AptSilentProgress()
            else:
                aptprogress = None

            apt_pkg.init_config()
            self.apt_cache = apt.Cache(progress=aptprogress)
        return self.apt_cache

    def lock(self) -> bool:
        """
        Lock the session to prevent further updates.

        @returns
          True if the session is locked
          False if another indexer is running
        """
        # Lock the session so that we prevent concurrent updates
        self.lockfd = os.open(axi.XAPIANDBLOCK, os.O_RDWR | os.O_CREAT)
        try:
            fcntl.lockf(self.lockfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            # Wrap the current progress with the server sender
            self.progress = ServerProgress(self.progress)
            return True
        except IOError as e:
            if e.errno == errno.EACCES or e.errno == errno.EAGAIN:
                return False
            else:
                raise

    def slave(self) -> None:
        """
        Attach to a running indexer and report its progress.

        Return when the other indexer has finished.
        """
        self.progress.notice("Another update is already running: showing its progress.")
        childProgress = ClientProgress(self.progress)
        childProgress.loop()

    def setupIndexing(self, force: bool = False, system: bool = True) -> bool:
        """
        Setup indexing: read plugins, check timestamps...

        @param force: if True, reindex also if the index is up to date

        @return:
          True if there is something to index
          False if there is no need of indexing
        """
        # Read values database
        #values = readValueDB(VALUESCONF, progress)

        # Read the plugins, in sorted order
        self.plugins = Plugins(progress=self.progress, system=system)

        # Ensure that we have something to do
        if len(self.plugins) == 0:
            self.progress.notice("No indexing plugins found in %s" % axi.PLUGINDIR)
            return False

        # Get the most recent modification timestamp of the data sources
        self.ds_timestamp = max([x.info['timestamp'] for x in self.plugins])

        # Get the timestamp of the last database update
        try:
            if os.path.exists(axi.XAPIANDBSTAMP):
                cur_timestamp = os.path.getmtime(axi.XAPIANDBSTAMP)
            else:
                cur_timestamp = 0
        except OSError as e:
            cur_timestamp = 0
            self.progress.notice("Reading current timestamp failed: %s. Assuming the index has not been created yet." % e)

        if self.verbose:
            self.progress.verbose("Most recent dataset:    %s." % time.ctime(self.ds_timestamp))
            self.progress.verbose("Most recent update for: %s." % time.ctime(cur_timestamp))

        # See if we need an update
        if cur_timestamp != 0 and int(self.ds_timestamp+.5) <= int(cur_timestamp+0.5):
            if force:
                self.progress.notice("The index %s is up to date, but rebuilding anyway as requested." % axi.XAPIANDBPATH)
            else:
                self.progress.notice("The index %s is up to date" % axi.XAPIANDBPATH)
                return False

        # Build the value database
        self.progress.verbose("Aggregating value information.")
        # Read existing value database to keep ids stable in a system
        self.values, self.values_desc = axi.readValueDB(quiet=True)
        values_seq = max(self.values.values()) + 1
        for addon in self.plugins:
            for v in addon.info.get("values", []):
                if v['name'] in self.values: continue
                self.values[v['name']] = values_seq
                values_seq += 1
                self.values_desc[v['name']] = v['desc']

        # Tell the plugins to do the long initialisation bits
        self.progress.verbose("Initializing plugins.")
        for addon in self.plugins:
            addon.obj.init(dict(values=self.values), self.progress)

        return True

    def get_document_from_apt(self, pkg: apt.Package) -> xapian.Document:
        """
        Get a xapian.Document for the given apt package record
        """
        assert self.plugins is not None
        document = xapian.Document()
        # The document data is the package name
        document.set_data(pkg.name)
        # add information about the version of the package in slot 0
        document.add_value(0, pkg.candidate.version)
        # Index the package name with a special prefix, to be able to find this
        # document by exact package name match
        document.add_term("XP"+pkg.name)
        # Have all the various plugins index their things
        for addon in self.plugins:
            addon.obj.index(document, pkg)
        return document

    def get_document_from_deb822(self, pkg) -> xapian.Document:
        """
        Get a xapian.Document for the given deb822 package record
        """
        assert self.plugins is not None
        document = xapian.Document()

        # The document data is the package name
        document.set_data(pkg["Package"])
        # add information about the version of the package in slot 0
        document.add_value(0, pkg["Version"])
        # Index the package name with a special prefix, to be able to find this
        # document by exact package name match
        document.add_term("XP"+pkg["Package"])
        # Have all the various plugins index their things
        for addon in self.plugins:
            addon.obj.indexDeb822(document, pkg)
        return document

    def gen_documents_apt(self) -> Iterator[xapian.Document]:
        """
        Generate Xapian documents from an apt cache
        """
        cache = self.aptcache()
        count = len(cache)
        for idx, pkg in enumerate(cache):
            if not pkg.candidate:
                continue
            # Print progress
            if idx % 200 == 0:
                self.progress.progress(100*idx/count)
            yield self.get_document_from_apt(pkg)

    def gen_documents_deb822(self, fnames: Iterable[str]) -> Iterator[xapian.Document]:
        from debian import deb822
        seen = set()
        for fname in fnames:
            infd = open(fname)
            # Get file size to compute progress
            total = os.fstat(infd.fileno())[6]
            for idx, pkg in enumerate(deb822.Deb822.iter_paragraphs(infd)):
                name = pkg["Package"]
                if name in seen:
                    continue
                seen.add(name)
                # Print approximate progress by checking the current read position
                # against the file size
                if total > 0 and idx % 200 == 0:
                    cur = infd.buffer.tell()
                    self.progress.progress(100*cur/total)
                yield self.get_document_from_deb822(pkg)

    def compareCacheToDb(
        self,
        cache: apt.Cache,
        db: xapian.WritableDatabase
    ) -> tuple[dict[str, int], dict[str, int], dict[str, int]]:
        """
        Compare the apt cache to the database and return dicts
        of the form (pkgname, docid) for the following states:

        unchanged - no new version since the last update
        outdated - a new version since the last update
        obsolete - no longer in the apt cache
        """
        unchanged: dict[str, int] = {}
        outdated: dict[str, int] = {}
        obsolete: dict[str, int] = {}
        self.progress.begin("Reading Xapian index")
        count = db.get_doccount()
        for (idx, m) in enumerate(db.postlist("")):
            if idx % 5000 == 0:
                self.progress.progress(100*idx/count)
            doc = db.get_document(m.docid)
            pkg = doc.get_data()
            if isinstance(pkg, bytes):
                # python3-apt requires strings.
                pkg = pkg.decode('utf-8')
            # this will return '' if there is no value 0, which is fine because it
            # will fail the comparison with the candidate version causing a reindex
            dbver = doc.get_value(0)
            # check if the package no longer exists
            if not cache.has_key(pkg) or not cache[pkg].candidate:
                obsolete[pkg] = m.docid
            # check if we have a new version, we do not have to delete
            # the record,
            elif cache[pkg].candidate.version != dbver:
                outdated[pkg] = m.docid
            # its a valid package and we know about it already
            else:
                unchanged[pkg] = m.docid
        self.progress.end()
        return unchanged, outdated, obsolete

    def updateIndex(self, pathname: str) -> None:
        """
        Update the index
        """
        assert self.plugins is not None
        db = xapian.WritableDatabase(pathname, xapian.DB_CREATE_OR_OPEN)
        cache = self.aptcache()
        count = len(cache)

        unchanged, outdated, obsolete = self.compareCacheToDb(cache, db)
        self.progress.verbose("Unchanged versions: %s, oudated version: %s, "
                         "obsolete versions: %s" % (len(unchanged),
                                                    len(outdated),
                                                    len(obsolete)))

        self.progress.begin("Updating Xapian index")
        for a in self.plugins:
            a.send_extra_info(db=db, aptcache=cache)
        for idx, pkg in enumerate(cache):
            if idx % 1000 == 0: self.progress.progress(100*idx/count)
            if not pkg.candidate:
                continue
            if pkg.name in unchanged:
                continue
            elif pkg.name in outdated:
                # update the existing
                db.replace_document(outdated[pkg.name], self.get_document_from_apt(pkg))
            else:
                # add the new ones
                db.add_document(self.get_document_from_apt(pkg))

        # and remove the obsoletes
        for docid in obsolete.values():
            db.delete_document(docid)

        # finished
        for a in self.plugins:
            a.finished()
        db.flush()
        self.progress.end()

    def incrementalUpdate(self) -> None:
        if not os.path.exists(axi.XAPIANINDEX):
            self.progress.notice("No Xapian index built yet: falling back to full rebuild")
            return self.rebuild()

        dbkind, dbpath = open(axi.XAPIANINDEX).readline().split()
        try:
            self.updateIndex(dbpath)
        except xapian.FeatureUnavailableError:
            self.progress.notice("Existing Xapian index is incompatible: falling back to full rebuild")
            return self.rebuild()

        # Update the index timestamp
        if not os.path.exists(axi.XAPIANDBSTAMP):
            open(axi.XAPIANDBSTAMP, "w").close()
        if self.ds_timestamp > 0:
            os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))

    def buildIndex(self, pathname: str, documents, addoninfo={}) -> None:
        """
        Create a new Xapian index with the content provided by the plugins
        """
        assert self.plugins is not None
        self.progress.begin("Rebuilding Xapian index")

        # Create a new Xapian index
        db = xapian.WritableDatabase(pathname, xapian.DB_CREATE_OR_OVERWRITE)
        # It seems to be faster without transactions, at the moment
        #db.begin_transaction(False)

        for a in self.plugins:
            a.send_extra_info(db=db)

        # Add all generated documents to the index
        for doc in documents:
            db.add_document(doc)

        #db.commit_transaction();
        for a in self.plugins:
            a.finished()
        db.flush()
        self.progress.end()

    def rebuild(self, pkgfiles: Iterable[str] | None = None) -> None:
        assert self.plugins is not None
        # Create a new Xapian index with the content provided by the plugins
        # Xapian takes care of preventing concurrent updates and removing the old
        # database if it's left over by a previous crashed update

        # Generate a new index name
        for idx in itertools.count(1):
            tmpidxfname = "index.%d" % idx
            dbdir = os.path.join(axi.XAPIANCACHEPATH, tmpidxfname)
            if not os.path.exists(dbdir): break;

        if pkgfiles:
            generator = self.gen_documents_deb822(pkgfiles)
        else:
            for a in self.plugins:
                a.send_extra_info(aptcache=self.aptcache())
            generator = self.gen_documents_apt()
        self.buildIndex(dbdir, generator)

        # Update the 'index' symlink to point at the new index
        self.progress.verbose("Installing the new index.")

        #os.symlink(tmpidxfname, axi.XAPIANDBPATH + "/index.tmp")
        with open(axi.XAPIANINDEX + ".tmp", "w") as out:
            print("auto", os.path.abspath(dbdir), file=out)
        os.rename(axi.XAPIANINDEX + ".tmp", axi.XAPIANINDEX)

        # Remove all other index.* directories that are not the newly created one
        def cleanoldcaches(dir: str) -> None:
            for file in os.listdir(dir):
                if not file.startswith("index."): continue
                # Don't delete what we just created
                if file == tmpidxfname: continue
                fullpath = os.path.join(dir, file)
                # Only delete directories
                if not os.path.isdir(fullpath): continue
                self.progress.verbose("Removing old index %s." % fullpath)
                shutil.rmtree(fullpath)
        cleanoldcaches(axi.XAPIANDBPATH)
        cleanoldcaches(axi.XAPIANCACHEPATH)

        # Commit the changes and update the last update timestamp
        if not os.path.exists(axi.XAPIANDBSTAMP):
            open(axi.XAPIANDBSTAMP, "w").close()
        if self.ds_timestamp > 0:
            os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))

        self.writeValues()
        self.writePrefixes()
        self.writeDoc()

    def writePrefixes(self, pathname: str = axi.XAPIANDBPREFIXES) -> None:
        """
        Write the prefix information on the given file
        """
        assert self.plugins is not None

        self.progress.verbose("Writing prefix information to %s." % pathname)
        with open(pathname+".tmp", "w") as out:
            print(textwrap.dedent("""
        # This file contains the information about keyword prefixes used in the
        # APT Xapian index.
        #
        # Xapian allows to prefix some terms so they can be told apart from
        # normal keywords: this is used for example with Debtags tags, and
        # stemmed forms.
        #
        # This file lists terms with their index prefix, their queryparser
        # prefix, whether queryparser should treat it as boolean or
        # probabilistic and a short description.
        """).lstrip(), file=out)

            # Aggregate and normalise prefix information from all the plugins
            prefixes: dict[str, dict[str, str]] = dict()
            for addon in self.plugins:
                for p in addon.info.get("prefixes", []):
                    idx = p.get("idx", None)
                    if idx is None: continue
                    qp = p.get("qp", None)
                    type = p.get("type", None)
                    desc = p.get("desc", None)
                    # TODO: warn of inconsistencies (plugins that disagree on qp or type)
                    old = prefixes.setdefault(idx, dict())
                    if qp: old.setdefault("qp", qp)
                    if type: old.setdefault("type", type)
                    if desc: old.setdefault("desc", desc)

            for name, info in sorted(prefixes.items(), key=itemgetter(0)):
                print("%s\t%s\t%s\t# %s" % (
                    name,
                    info.get("qp", "-"),
                    info.get("type", "-"),
                    info.get("desc", "(description is missing)")), file=out)

        # Atomic update of the documentation
        os.rename(pathname+".tmp", pathname)

    def writeValues(self, pathname: str = axi.XAPIANDBVALUES) -> None:
        """
        Write the value information on the given file
        """
        self.progress.verbose("Writing value information to %s." % pathname)
        with open(pathname+".tmp", "w") as out:

            print(textwrap.dedent("""
        # This file contains the mapping between names of numeric values indexed in the
        # APT Xapian index and their index
        #
        # Xapian allows to index numeric values as well as keywords and to use them for
        # all sorts of useful querying tricks.  However, every numeric value needs to
        # have a unique index, and this configuration file is needed to record which
        # indices are allocated and to provide a mnemonic name for them.
        #
        # The format is exactly like /etc/services with name, number and optional
        # aliases, with the difference that the second column does not use the
        # "/protocol" part, which would be meaningless here.
        """).lstrip(), file=out)

            for name, idx in sorted(self.values.items(), key=itemgetter(1)):
                desc = self.values_desc[name]
                print("%s\t%d\t# %s" % (name, idx, desc), file=out)

        # Atomic update of the documentation
        os.rename(pathname+".tmp", pathname)

    def writeDoc(self, pathname: str = axi.XAPIANDBDOC) -> None:
        """
        Write the documentation in the given file
        """
        assert self.plugins is not None
        self.progress.verbose("Writing documentation to %s." % pathname)
        # Collect the documentation
        docinfo = []
        for addon in self.plugins:
            try:
                doc = addon.obj.doc()
                if doc != None:
                    docinfo.append(dict(
                        name = doc['name'],
                        shortDesc = doc['shortDesc'],
                        fullDoc = doc['fullDoc']))
            except:
                # If a plugin has problem returning documentation, don't worry about it
                self.progress.notice("Skipping documentation for plugin", addon.filename)

        # Write the documentation in pathname
        with open(pathname+".tmp", "w") as out:
            print(textwrap.dedent("""
        ===============
        Database layout
        ===============

        This Xapian database indexes Debian package information.  To query the
        database, open it as ``%s/index``.

        Data are indexed either as terms or as values.  Words found in package
        descriptions are indexed lowercase, and all other kinds of terms have an
        uppercase prefix as documented below.

        Numbers are indexed as Xapian numeric values.  A list of the meaning of the
        numeric values is found in ``%s``.

        The data sources used for indexing are:
        """).lstrip() % (axi.XAPIANDBPATH, axi.XAPIANDBVALUES), file=out)

            for d in docinfo:
                print(" * %s: %s" % (d['name'], d['shortDesc']), file=out)

                print(textwrap.dedent("""
        This Xapian index follows the conventions for term prefixes described in
        ``/usr/share/doc/xapian-omega/termprefixes.txt.gz``.

        Extra Debian data sources can define more extended prefixes (starting with
        ``X``): their meaning is documented below together with the rest of the data
        source documentation.

        At the very least, at least the package name (with the ``XP`` prefix) will
        be present in every document in the database.  This allows to quickly
        lookup a Xapian document by package name.

        The user data associated to a Xapian document is the package name.


        -------------------
        Active data sources
        -------------------

        """), file=out)
            for d in docinfo:
                print(d['name'], file=out)
                print('='*len(d['name']), file=out)
                print(textwrap.dedent(d['fullDoc']), file=out)
                print(file=out)

        # Atomic update of the documentation
        os.rename(pathname+".tmp", pathname)




#!/usr/bin/python3

# axi-query-adaptivecutoff - Use an adaptive cutoff to select results
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options] keywords",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")
parser.add_option("-t", "--type", help="package type, one of 'game', 'gui', 'cmdline' or 'editor'")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import xapian
from aptxapianindex import (
    XAPIANDB,
    termsForSimpleQuery,
    addSimpleFilterToQuery,
    show_mset,
)

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Build the base query as seen in axi-query-simple.py
query = xapian.Query(xapian.Query.OP_OR, termsForSimpleQuery(args))

# Add the simple user filter, if requeste
query = addSimpleFilterToQuery(query, options.type)

# Perform the query
enquire = xapian.Enquire(db)
enquire.set_query(query)

# Retrieve the first result, and check its relevance
matches = enquire.get_mset(0, 1)
topWeight = matches[0].weight

# Tell Xapian that we only want results that are at least 70% as good as that
enquire.set_cutoff(0, topWeight * 0.7)

# Now we have a meaningful cutoff, so we can get a larger number of results:
# thanks to the cutoff, approximate results will stop before starting to be
# really bad.
matches = enquire.get_mset(0, 200)

# Display the results
show_mset(matches)

sys.exit(0)



#!/usr/bin/python3

# axi-query-tags - Look for Debtags tags by keyword
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options] keywords",
                version="%prog "+ VERSION,
                description="Find the Debtags tags corresponding to some keywords")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import xapian
from aptxapianindex import (
    XAPIANDB,
    termsForSimpleQuery,
)

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Build the base query
query = xapian.Query(xapian.Query.OP_OR, termsForSimpleQuery(args))

# Perform the query
enquire = xapian.Enquire(db)
enquire.set_query(query)

# Now, instead of showing the results of the query, we ask Xapian what are the
# terms in the index that are most relevant to this search.
# Normally, you would use the results to suggest the user possible ways for
# refining the search.  I instead abuse this feature to see what are the tags
# that are most related to the search results.

# Use an adaptive cutoff to avoid to pick bad results as references
matches = enquire.get_mset(0, 1)
topWeight = matches[0].weight
enquire.set_cutoff(0, topWeight * 0.7)

# Select the first 10 documents as the key ones to use to compute relevant
# terms
rset = xapian.RSet()
for m in enquire.get_mset(0, 10):
    rset.add_document(m.docid)

# Xapian supports providing a filter object, to say that we are only interested
# in some terms.
# This one filters out all the keywords that are not tags, or that were in the
# list of query terms.
class Filter(xapian.ExpandDecider):
    def __call__(self, term):
        """
        Return true if we want the term, else false
        """
        return term[:2] == "XT"

# This is the "Expansion set" for the search: the 10 most relevant terms that
# match the filter
eset = enquire.get_eset(10, rset, Filter())

# Print out the results
for res in eset:
    print("%.2f %s" % (res.weight, res.term.decode('utf-8')[2:]))

sys.exit(0)



#!/usr/bin/python3

# axi-searchasyoutype - Search-as-you-type demo
#
# Copyright (C) 2008  Matteo Zandi, Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Note: this program needs python-gtkhtml2 to work

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options] keywords",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")
parser.add_option("-t", "--type", help="package type, one of 'game', 'gui', 'cmdline' or 'editor'")

(options, args) = parser.parse_args()

import math
# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import gtk
import gtkhtml2
import xapian
from aptxapianindex import (
    XAPIANDB,
    termsForSimpleQuery,
    addSimpleFilterToQuery,
)

from debian import deb822

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Instantiate the APT cache as well
import apt
cache = apt.Cache()

# Facet name -> Short description
facets = dict()
# Tag name -> Short description
tags = dict()
for p in deb822.Deb822.iter_paragraphs(open("/var/lib/debtags/vocabulary", "r")):
    if "Description" not in p: continue
    desc = p["Description"].split("\n", 1)[0]
    if "Tag" in p:
        tags[p["Tag"]] = desc
    elif "Facet" in p:
        facets[p["Facet"]] = desc

def SimpleOrQuery(input_terms):
    """
    Simple OR Query

    terms is an array of words
    """

    if len(input_terms) == 0:
        # No text given: abort
        return []

    # To understand the following code, keep in mind that we do
    # search-as-you-type, so the last word may be partially typed.

    if len(input_terms[-1]) == 1:
        # If the last term has only one character, we ignore it waiting for
        # the user to type more.  A better implementation can set up a
        # timer to disable this some time after the user stopped typing, to
        # give meaningful results to searches like "programming in c"
        input_terms = input_terms[:-1]
        if len(input_terms) == 0:
            return [], []

    # Convert the words into terms for the query
    terms = termsForSimpleQuery(input_terms)

    # Since the last word can be partially typed, we add all words that
    # begin with the last one.
    terms.extend([x.term.decode('utf-8') for x in db.allterms(input_terms[-1])])

    # Build the query
    query = xapian.Query(xapian.Query.OP_OR, terms)

    # Add the simple user filter, if requested.  This is to show that even
    # when we do search-as-you-type, everything works as usual
    query = addSimpleFilterToQuery(query, options.type)

    # Perform the query
    enquire = xapian.Enquire(db)
    enquire.set_query(query)

    # This does the adaptive cutoff trick on the query results (see
    # axi-query-adaptivecutoff.py) 
    mset = enquire.get_mset(0, 1)
    try:
        topWeight = mset[0].weight
    except IndexError:
        return [], []
    enquire.set_cutoff(0, topWeight * 0.7)

    # Select the first 30 documents as the key ones to use to compute relevant
    # terms
    packages = []
    rset = xapian.RSet()
    for m in enquire.get_mset(0, 30):
        rset.add_document(m.docid)
        name = m.document.get_data()
        score = m.percent
        try:
            pkg = cache[name]
        except KeyError:
            continue

        # pkg.candidate may be none
        try:
            shortdesc = pkg.candidate.summary
        except AttributeError:
            continue

        packages.append((score, name, shortdesc))

    class Filter(xapian.ExpandDecider):
        def __call__(self, term):
            #return term[0].islower() or term[:2] == "XT"
            return term[:2] == "XT"

    def format(k):
        if k in tags:
            facet = k.split("::", 1)[0]
            if facet in facets:
                return "<i>%s: %s</i>" % (facets[facet], tags[k])
            else:
                return "<i>%s</i>" % tags[k]
        else:
            return k

    taglist = []
    maxscore = None
    for res in enquire.get_eset(15, rset, Filter()):
        # Normalise the score in the interval [0, 1]
        weight = math.log(res.weight)
        if maxscore == None: maxscore = weight
        tag = res.term.decode('utf-8')[2:]
        taglist.append(
            (tag, format(tag), float(weight) / maxscore)
        )
    taglist.sort(key=lambda x:x[0])

    if len(taglist) == 0:
        return [], []

    return packages, taglist

def mark_text_up(result_list):
    # 0-100 score, key (facet::tag), description
    document = gtkhtml2.Document()
    document.clear()
    document.open_stream("text/html")
    document.write_stream("""<html><head>
<style type="text/css">
a { text-decoration: none; color: black; }
</style>
</head><body>""")
    for tag, desc, score in result_list:
        document.write_stream('<a href="%s" style="font-size: %d%%">%s</a> ' % (tag, score*150, desc))
    document.write_stream("</body></html>")
    document.close_stream()
    return document

class Demo:
    def __init__(self):
        w = gtk.Window()
        w.connect('destroy', gtk.main_quit)
        w.set_default_size(400, 400)

        self.model = gtk.ListStore(int, str, str)

        treeview = gtk.TreeView()
        treeview.show()
        treeview.set_model(self.model)

        cell_pct = gtk.CellRendererText()
        column_pct = gtk.TreeViewColumn ("Percent", cell_pct, text=0)
        column_pct.set_sort_column_id(0)
        treeview.append_column(column_pct)

        cell_name = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn ("Name", cell_name, text=1)
        column_name.set_sort_column_id(0)
        treeview.append_column(column_name)

        cell_desc = gtk.CellRendererText()
        column_desc = gtk.TreeViewColumn ("Name", cell_desc, text=2)
        column_desc.set_sort_column_id(0)
        treeview.append_column(column_desc)

        document = gtkhtml2.Document()
        document.clear()
        document.open_stream("text/html")
        document.write_stream("<html><body>Welcome, enter some text to start!</body></html>")
        document.close_stream()
        self.view = gtkhtml2.View()
        self.view.set_document(document)

        scrolledwin = gtk.ScrolledWindow()
        scrolledwin.show()
        scrolledwin.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        scrolledwin.add(treeview)

        vbox = gtk.VBox(False, 0)
        vbox.pack_start(scrolledwin, True, True, 0)
        vbox.pack_start(self.view, True, True, 0)

        self.entry = gtk.Entry()
        self.entry.connect('changed', self.on_entry_changed)
        vbox.pack_start(self.entry, False, True, 0)

        w.add(vbox)
        w.show_all()
        gtk.main()

    def on_entry_changed(self, widget, *args):
        packageslist, tags = SimpleOrQuery(widget.get_text().split())

        self.model.clear()
        for item in packageslist:
            self.model.append(item)

        doc = mark_text_up(tags)
        doc.connect('link_clicked', self.on_link_clicked)
        self.view.set_document(doc)

    def on_link_clicked(self, document, link):
    	self.entry.set_text(link + " " + self.entry.get_text().lstrip())

if __name__ == "__main__":
    demo = Demo()



#!/usr/bin/python3

import dbus
import dbus.mainloop.glib
import glib

dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

def finished(res):
    print("finished: ", res)

def progress(percent):
    print("progress: ", percent)

system_bus = dbus.SystemBus()

axi = dbus.Interface(system_bus.get_object("org.debian.AptXapianIndex","/"),
                     "org.debian.AptXapianIndex")
axi.connect_to_signal("UpdateFinished", finished)
axi.connect_to_signal("UpdateProgress", progress)
# force, update_only
axi.update_async(True, True)

glib.MainLoop().run()




try:
    from xdg import Locale
    from xdg.DesktopEntry import DesktopEntry
    from xdg.Exceptions import ParsingError
    HAS_XDG=True
except ImportError as e:
    HAS_XDG=False

import os
import os.path

import xapian

APPINSTALLDIR="/usr/share/app-install/desktop/"

class Indexer:
    def __init__(self, lang, val_popcon, progress=None):
        self.val_popcon = val_popcon
        self.progress = progress
        if lang is None:
            lang = "en"
        self.lang = lang
        self.xlang = lang.split("_")[0]
        self.xdglangs = Locale.expand_languages([lang])
        self.indexer = xapian.TermGenerator()
        # Get a stemmer for this language, if available
        try:
            self.stemmer = xapian.Stem(self.xlang)
            self.indexer.set_stemmer(self.stemmer)
        except xapian.InvalidArgumentError:
            pass

    def index(self, document, fname, entry):
        # Index a single term "XD", marking that the package contains .desktop
        # files
        document.add_term("XD")

        # Index the name of the .desktop file, with prefix XDF
        document.add_term("XDF" + fname)

        # Index keywords retrieved in this indexer's language
        self.indexer.set_document(document)
        oldlangs = Locale.langs
        try:
            Locale.langs = self.xdglangs
            self.indexer.index_text_without_positions(entry.getName())
            self.indexer.index_text_without_positions(entry.getGenericName())
            self.indexer.index_text_without_positions(entry.getComment())
        finally:
            Locale.langs = oldlangs

        # Index .desktop categories, with prefix XDT
        for cat in entry.getCategories():
            document.add_term("XDT"+cat)

        # Add an "app-popcon" value with popcon rank
        try:
            popcon = int(entry.get("X-AppInstall-Popcon"))
        except ValueError as e:
            if self.progress:
                self.progress.verbose("%s: parsing X-AppInstall-Popcon: %s" % (fname, str(e)))
            popcon = -1
        if self.val_popcon != -1:
            document.add_value(self.val_popcon, xapian.sortable_serialise(popcon));


class AppInstall(object):
    def __init__(self, langs, progress):
        self.langs = langs
        self.progress = progress

    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:

         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        maxts = 0
        for f in os.listdir(APPINSTALLDIR):
            if f[0] == '.' or not f.endswith(".desktop"): continue
            ts = os.path.getmtime(os.path.join(APPINSTALLDIR, f))
            if ts > maxts: maxts = ts
        return dict(
                timestamp = maxts,
                values = [
                    dict(name = "app-popcon", desc = "app-install .desktop popcon rank"),
                ],
                sources = [
                    dict(path=APPINSTALLDIR, desc=".desktop files provided by app-install-data"),
                ],
                prefixes = [
                    dict(idx="XD", qp=None, type=None,
                         desc="Marker to indicate that the package contains .desktop files",
                         ldesc="Only 'XD' can present in the index. This is used to efficiently"
                               " filter packages that have a .desktop file"),
                    dict(idx="XDF", qp=None, type=None,
                         desc="File name of the .desktop file",
                         ldesc="This is the name of a .desktop file contained in the package."
                               " There could be more than one"),
                    dict(idx="XDT", qp=None, type="bool",
                         desc="Categories from .desktop files",
                         ldesc="This is similar to a tag, but filled with the categories found"
                               " in .desktop files"),
                    dict(idx="Z", qp=None, type=None,
                         desc="Stemmed forms of keywords",
                         ldesc="This contains the stemmed forms of keywords as generated by"
                               " TermGenerator and matched by QueryParser"),
                ])

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """

        # Read the value indexes we will use
        values = info['values']
        self.val_popcon = values.get("app-popcon", -1)

        self.indexers = [Indexer(lang, self.val_popcon, progress) for lang in [None] + list(self.langs)]
        self.entries = {}

        progress.begin("Reading .desktop files from %s" % APPINSTALLDIR)
        for f in os.listdir(APPINSTALLDIR):
            if f[0] == '.' or not f.endswith(".desktop"): continue
            try:
                entry = DesktopEntry(os.path.join(APPINSTALLDIR, f))
            except (ValueError, ParsingError, UnicodeDecodeError):
                # Invalid .desktop files can cause a ValueError. From PyXDG 0.25,
                # that case will be turned into a ParsingError.
                continue
            pkg = entry.get("X-AppInstall-Package")
            self.entries.setdefault(pkg, []).append((f, entry))
        progress.end()

    def send_extra_info(self, db=None, **kw):
        """
        Receive extra parameters from the indexer.

        This may be called more than once, but after init().

        We are using this to get the database instance
        """
        if db is not None:
            for i in self.indexers:
                i.indexer.set_flags(xapian.TermGenerator.FLAG_SPELLING)
                i.indexer.set_database(db)

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        return dict(
            name = "app-install information",
            shortDesc = "terms, categories and popcon values extracted from the app-install .desktop files",
            fullDoc = """
            The AppInstall data source reads .desktop files from %s
            and adds the following terms:

             * keywords from the .desktop descriptions, via Xapian's
               TermGenerator, in all requested locales;
             * .desktop categories, with prefix XDT;
             * name of .desktop file, with prefix XDF;
             * a single term "XD", marking that the package contains .desktop
               files.

            It also adds an "app-popcon" value with popcon ranks from the
            app-install .desktop files.
            """ % APPINSTALLDIR
        )

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        name = document.get_data()
        for e in self.entries.get(name, []):
            fname, entry = e
            for i in self.indexers:
                i.index(document, fname, entry)

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        name = document.get_data()
        for e in self.entries.get(name, []):
            fname, entry = e
            for i in self.indexers:
                i.index(document, fname, entry)

def init(langs=None, progress=None, **kw):
    """
    Create and return the plugin object.
    """
    # If we don't have app-install data, skip it
    if not os.path.isdir(APPINSTALLDIR): return None
    # If we don't have python-xdg, skip it
    if not HAS_XDG:
        if progress:
            progress.verbose("please install python-xdg if you want to index app-install-data files")
        return None
    return AppInstall(langs=langs, progress=progress)



#!/usr/bin/python3

# axi-searchasyoutype - Search-as-you-type demo
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options] keywords",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")
parser.add_option("-t", "--type", help="package type, one of 'game', 'gui', 'cmdline' or 'editor'")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import xapian
from aptxapianindex import (
    XAPIANDB,
    termsForSimpleQuery,
    addSimpleFilterToQuery,
)

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Instantiate the APT cache as well
import apt
cache = apt.Cache()

import curses
import curses.wrapper
import re


class Results:
    """
    Show the results of a query while we type it
    """
    def __init__(self, stdscr):
        maxy, maxx = stdscr.getmaxyx()
        self.size = maxy-1
        self.splitline = re.compile(r'\s+')
        self.win = curses.newwin(self.size, maxx, 0, 0)
        self.win.clear()

    def noresults(self, suggestion = "type more"):
        self.win.clear()
        self.win.addstr(0, 0, "No results, " + suggestion, curses.A_BOLD)
        self.win.refresh()

    def update(self, line):
        """
        Show the results of the search done using the given line of text
        """
        line = line.lower().strip()
        if len(line) == 0:
            # No text given: abort
            self.noresults()
            return

        # Split the line in words
        args = self.splitline.split(line)

        if len(args) == 0:
            # No text given: abort
            self.noresults()
            return

        # To understand the following code, keep in mind that we do
        # search-as-you-type, so the last word may be partially typed.

        if len(args[-1]) == 1:
            # If the last term has only one character, we ignore it waiting for
            # the user to type more.  A better implementation can set up a
            # timer to disable this some time after the user stopped typing, to
            # give meaningful results to searches like "programming in c"
            args = args[:-1]
            if len(args) == 0:
                self.noresults()
                return

        # Convert the words into terms for the query
        terms = termsForSimpleQuery(args)

        # Since the last word can be partially typed, we add all words that
        # begin with the last one.
        terms.extend([x.term.decode('utf-8') for x in db.allterms(args[-1])])

        # Build the query
        query = xapian.Query(xapian.Query.OP_OR, terms)

        # Add the simple user filter, if requested.  This is to show that even
        # when we do search-as-you-type, everything works as usual
        query = addSimpleFilterToQuery(query, options.type)

        # Perform the query
        enquire = xapian.Enquire(db)
        enquire.set_query(query)

        # This does the adaptive cutoff trick on the query results (see
        # axi-query-adaptivecutoff.py) 
        mset = enquire.get_mset(0, 1)
        try:
            topWeight = mset[0].weight
        except IndexError:
            self.noresults("change your query")
            return
        enquire.set_cutoff(0, topWeight * 0.7)

        # Retrieve as many results as we can show
        mset = enquire.get_mset(0, self.size - 1)

        # Redraw the window
        self.win.clear()

        # Header
        self.win.addstr(0, 0, "%i results found." % mset.get_matches_estimated(), curses.A_BOLD)

        # Results
        for y, m in enumerate(mset):
            # /var/lib/apt-xapian-index/README tells us that the Xapian document data
            # is the package name.
            name = m.document.get_data()

            # Get the package record out of the Apt cache, so we can retrieve the short
            # description
            pkg = cache[name]

            if pkg.candidate:
                # Print the match, together with the short description
                self.win.addstr(y+1, 0, "%i%% %s - %s" % (m.percent, name, pkg.candidate.summary))

        self.win.refresh()

# Build the base query as seen in axi-query-simple.py

class Input:
    def __init__(self, stdscr, results):
        maxy, maxx = stdscr.getmaxyx()
        self.results = results
        self.win = curses.newwin(1, maxx, maxy-1, 0)
        self.win.bkgdset(' ', curses.A_REVERSE)
        self.win.clear()
        self.win.addstr(0, 0, "> ", curses.A_REVERSE)
        self.line = ""

    def mainloop(self):
        old = ""
        while True:
            c = self.win.getch()
            if c == 10 or c == 27:
                break
            elif c == 127:
                control = True
                if len(self.line) > 0:
                    self.line = self.line[:-1]
            else:
                self.line += chr(c)
            self.win.clear()
            self.win.addstr(0, 0, "> " + self.line, curses.A_REVERSE)
            self.win.refresh()
            if self.line != old:
                self.results.update(self.line)
                old = self.line

def main(stdscr):
    results = Results(stdscr)
    input = Input(stdscr, results)
    stdscr.refresh()
    input.mainloop()

curses.wrapper(main)

sys.exit(0)



#
# axi - apt-xapian-index python modules
#
# Copyright (C) 2007--2010  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import os.path
import re
import sys

# Setup configuration
PLUGINDIR = os.environ.get("AXI_PLUGIN_DIR", "/usr/share/apt-xapian-index/plugins")
XAPIANDBPATH = os.environ.get("AXI_DB_PATH", "/var/lib/apt-xapian-index")
XAPIANDBSTAMP = os.path.join(XAPIANDBPATH, "update-timestamp")
XAPIANDBLOCK = os.path.join(XAPIANDBPATH, "update-lock")
XAPIANDBUPDATESOCK = os.path.join(XAPIANDBPATH, "update-socket")
XAPIANDBVALUES = os.path.join(XAPIANDBPATH, "values")
XAPIANDBPREFIXES = os.path.join(XAPIANDBPATH, "prefixes")
XAPIANDBDOC = os.path.join(XAPIANDBPATH, "README")
XAPIANINDEX = os.path.join(XAPIANDBPATH, "index")
XAPIANCACHEPATH = os.environ.get("AXI_CACHE_PATH", "/var/cache/apt-xapian-index")

# Default value database in case one cannot be read
DEFAULT_VALUES = dict(version=0, installedsize=1, packagesize=2)
DEFAULT_VALUE_DESCS = dict(
        version="package version",
        installedsize="installed size",
        packagesize="package size"
)

def readValueDB(
    pathname: str = XAPIANDBVALUES,
    quiet: bool = False
) -> tuple[dict[str, int], dict[str, str]]:
    """
    Read the "/etc/services"-style database of value indices
    """
    try:
        re_empty = re.compile(r"^\s*(?:#.*)?$")
        re_value = re.compile(r"^(\S+)\s+(\d+)(?:\s*#\s*(.+))?$")
        values: dict[str, int] = {}
        descs: dict[str, str] = {}
        for idx, line in enumerate(open(pathname)):
            # Skip empty lines and comments
            if re_empty.match(line): continue
            # Parse teh rest
            mo = re_value.match(line)
            if not mo:
                if not quiet:
                    print("%s:%d: line is not `name value [# description]': ignored" % (pathname, idx+1), file=sys.stderr)
                continue
            # Parse the number
            name = mo.group(1)
            number = int(mo.group(2))
            desc = mo.group(3) or ""

            values[name] = number
            descs[name] = desc
        if not values:
            raise BrokenIndexError
    except (OSError, IOError, BrokenIndexError) as e:
        # If we can't read the database, fallback to defaults
        if not quiet:
            print("%s: %s. Falling back on a default value database" % (pathname, e), file=sys.stderr)
        values = DEFAULT_VALUES
        descs = DEFAULT_VALUE_DESCS
    return values, descs


class BrokenIndexError(Exception):
    pass



# -*- coding: utf-8 -*-
from __future__ import annotations

import os.path
import unittest

import axi
import axi.indexer


class AxiTestBase(unittest.TestCase):
    def assertCleanIndex(self) -> None:
        self.assert_(os.path.exists(axi.XAPIANINDEX))
        self.assert_(os.path.exists(axi.XAPIANDBSTAMP))
        self.assert_(os.path.exists(axi.XAPIANDBVALUES))
        self.assert_(os.path.exists(axi.XAPIANDBDOC))
        self.assert_(not os.path.exists(axi.XAPIANDBUPDATESOCK))
        indexer: axi.indexer.Indexer | None

        # Index is clean and up to date: an indexer should tell us that it does
        # not need to run
        progress = axi.indexer.SilentProgress()
        indexer = axi.indexer.Indexer(progress, True)
        self.assert_(indexer.lock())
        self.assert_(not indexer.setupIndexing())
        indexer = None



#!/usr/bin/python3

# axi-query-pkgtype - Like axi-query-simple.py, but with a simple
#                     result filter
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options]",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")
parser.add_option("-t", "--type", help="package type, one of 'game', 'gui', 'cmdline' or 'editor'")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import warnings

import xapian
from aptxapianindex import (
    XAPIANDB,
    termsForSimpleQuery,
)

# This tells python-apt that we've seen the warning about the API not being
# stable yet, and we don't want to see every time we run the program
warnings.filterwarnings("ignore","apt API not stable yet")
import apt

warnings.resetwarnings()

# This is our little database of simple Debtags filters we provide: the name
# entered by the user in "--type" maps to a piece of Xapian query
filterdb = dict(
    # We can do simple AND queries...
    game = xapian.Query(xapian.Query.OP_AND, ('XTuse::gameplaying', 'XTrole::program')),
    # Or we can do complicate binary expressions...
    gui = xapian.Query(xapian.Query.OP_AND, xapian.Query('XTrole::program'),
                xapian.Query(xapian.Query.OP_OR, 'XTinterface::x11', 'XTinterface::3d')),
    cmdline = xapian.Query(xapian.Query.OP_AND, 'XTrole::program', 'XTinterface::commandline'),
    editor = xapian.Query(xapian.Query.OP_AND, 'XTrole::program', 'XTuse::editing'))
    # Feel free to invent more

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Build the base query as seen in axi-query-simple.py
query = xapian.Query(xapian.Query.OP_OR, termsForSimpleQuery(args))

# See if the user wants to use one of the result filters
if options.type:
    if options.type in filterdb:
        # If a filter was requested, AND it with the query
        query = xapian.Query(xapian.Query.OP_AND, query, filterdb[options.type])
    else:
        print("Invalid filter type.  Try one of", ", ".join(sorted(filterdb)), file=sys.stderr)
        sys.exit(1)


# Perform the query, all the rest is as in axi-query-simple.py
enquire = xapian.Enquire(db)
enquire.set_query(query)

# Display the top 20 results, sorted by how well they match
cache = apt.Cache()
matches = enquire.get_mset(0, 20)
print("%i results found." % matches.get_matches_estimated())
print("Results 1-%i:" % matches.size())
for m in matches:
    # /var/lib/apt-xapian-index/README tells us that the Xapian document data
    # is the package name.
    name = m.document.get_data()

    # Get the package record out of the Apt cache, so we can retrieve the short
    # description
    pkg = cache[name]

    if pkg.candidate:
    # Print the match, together with the short description
        print("%i%% %s - %s" % (m.percent, name, pkg.candidate.summary))

sys.exit(0)



try:
    import apt_pkg
    HAS_APT=True
except ImportError:
    HAS_APT=False
import os
import os.path
import pickle
import time

import xapian


class CatalogedTime:
    
    PERSISTANCE_DIR = "/var/lib/apt-xapian-index/"
    CATALOGED_NAME = "cataloged_times.p"

    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:

         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        res = dict(timestamp=0,
                   values=[
                       dict(name="catalogedtime", desc="Cataloged timestamp"),
                       ],
                   sources=[
                       dict(
                           path=os.path.join(self.PERSISTANCE_DIR, self.CATALOGED_NAME),
                           desc="first-seen information for every package"
                       )
                   ])
        if not HAS_APT: return res
        if not hasattr(apt_pkg, "config"): return res
        fname = apt_pkg.config.find_file("Dir::Cache::pkgcache")
        if not os.path.exists(fname): return res
        res["timestamp"] = os.path.getmtime(fname)
        return res

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """
        if os.access(self.PERSISTANCE_DIR, os.W_OK):
            if not os.path.exists(self.PERSISTANCE_DIR):
                os.makedirs(self.PERSISTANCE_DIR)
            self._packages_cataloged_file = os.path.join(self.PERSISTANCE_DIR, 
                                                         self.CATALOGED_NAME)
        else:
            self._packages_cataloged_file = None
        if (self._packages_cataloged_file and 
            os.path.exists(self._packages_cataloged_file)):
            self._package_cataloged_time = pickle.load(open(self._packages_cataloged_file, "rb"))
        else:
            self._package_cataloged_time = {}
        self.now = time.time()
        values = info['values']
        self.value = values.get("catalogedtime", -1)

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        return dict(
            name = "Cataloged time",
            shortDesc = "store the timestamp when the package was first cataloged",
            fullDoc = """
            This datasource simply stores a value with the timestamp
            when the package was first cataloged. This is useful to e.g.
            implement a 'Whats new' feature.
            """
        )

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        time = self._package_cataloged_time.get(pkg.name, self.now)
        self._package_cataloged_time[pkg.name] = time
        document.add_value(self.value, xapian.sortable_serialise(time))

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        pass

    def finished(self):
        """
        Called when the indexing is finihsed
        """
        if self._packages_cataloged_file:
            f=open(self._packages_cataloged_file+".new", "wb")
            res = pickle.dump(self._package_cataloged_time, f)
            f.close()
            os.rename(self._packages_cataloged_file+".new",
                      self._packages_cataloged_file)

def init(**kw):
    """
    Create and return the plugin object.
    """
    if not HAS_APT: return None
    return CatalogedTime()



#!/usr/bin/python3

# axi-query-expand - Query and show possible expansions
#
# Copyright (C) 2007  Enrico Zini <enrico@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from optparse import OptionParser

VERSION="0.1"

# Let's start with a simple command line parser with help
class Parser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)

    def error(self, msg):
        sys.stderr.write("%s: error: %s\n\n" % (self.get_prog_name(), msg))
        self.print_help(sys.stderr)
        sys.exit(2)

parser = Parser(usage="usage: %prog [options] keywords",
                version="%prog "+ VERSION,
                description="Query the Apt Xapian index.  Command line arguments can be keywords or Debtags tags")
parser.add_option("-t", "--type", help="package type, one of 'game', 'gui', 'cmdline' or 'editor'")

(options, args) = parser.parse_args()


# Import the rest here so we don't need dependencies to be installed only to
# print commandline help
import xapian
from aptxapianindex import (
    XAPIANDB,
    termsForSimpleQuery,
    addSimpleFilterToQuery,
    show_mset,
)

# Instantiate a xapian.Database object for read only access to the index
db = xapian.Database(XAPIANDB)

# Build the base query as seen in axi-query-simple.py
query = xapian.Query(xapian.Query.OP_OR, termsForSimpleQuery(args))

# Add the simple user filter, if requeste
query = addSimpleFilterToQuery(query, options.type)

# Perform the query
enquire = xapian.Enquire(db)
enquire.set_query(query)

# Retrieve the top 20 results
matches = enquire.get_mset(0, 20)

# Display the results
show_mset(matches)

# Now, we ask Xapian what are the terms in the index that are most relevant to
# this search.  This can be used to suggest to the user the most useful ways of
# refining the search.

# Select the first 10 documents as the key ones to use to compute relevant
# terms
rset = xapian.RSet()
for m in matches:
    rset.add_document(m.docid)

# This is the "Expansion set" for the search: the 10 most relevant terms
eset = enquire.get_eset(10, rset)

# Print it out.  Note that some terms have a prefix from the database: can we
# filter them out?  Indeed: Xapian allow to give a filter to get_eset.
# Read on...
print()
print("Terms that could improve the search:", end='')
print(", ".join(["%s (%.2f%%)" % (res.term.decode('utf-8'), res.weight) for res in eset]))


# You can also abuse this feature to show what are the tags that are most
# related to the search results.  This allows you to turn a search based on
# keywords to a search based on semantic attributes, which would be an
# absolutely stunning feature in a GUI.

# We can do it thanks to Xapian allowing to specify a filter for the output of
# get_eset.  This filter filters out all the keywords that are not tags, or
# that were in the list of query terms.
class Filter(xapian.ExpandDecider):
    def __call__(self, term):
        """
        Return true if we want the term, else false
        """
        return term[:2] == "XT"

# This is the "Expansion set" for the search: the 10 most relevant terms that
# match the filter
eset = enquire.get_eset(10, rset, Filter())

# Print out the resulting tags
print()
print("Tags that could improve the search:", end='')
print(", ".join(["%s (%.2f%%)" % (res.term.decode('utf-8')[2:], res.weight) for res in eset]))

sys.exit(0)



# Add debtags tags to the index

import os
import os.path
import re

from debian import debtags

DEBTAGSDB = os.environ.get("AXI_DEBTAGS_DB", "/var/lib/debtags/package-tags")

class Debtags:
    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:

         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        return dict(
                timestamp=os.path.getmtime(DEBTAGSDB),
                sources=[dict(path=DEBTAGSDB, desc="Debtags tag information")],
                prefixes=[
                    dict(idx="XT", qp="tag:", type="bool",
                         desc="Debtags tag",
                         ldesc="Debtags package categories"),
                ],
        )

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """
        progress.begin("Reading Debtags database")
        self.db = debtags.DB()
        # Read full database, skipping special and todo tags
        tagFilter = re.compile(r"^special::.+$|^.+::TODO$")
        self.db.read(open(DEBTAGSDB, "r"), lambda x: not tagFilter.match(x))
        progress.end()

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        return dict(
            name = "Debtags",
            shortDesc = "Debtags tag information",
            fullDoc = """
            The Debtags data source indexes Debtags tags as terms with the
            ``XT`` prefix; for example: 'XTrole::program'.

            Using the ``XT`` terms, queries can be enhanced with semantic
            information.  Xapian's support for complex expressions in queries
            can be used to great effect: for example::

                XTrole::program AND XTuse::gameplaying AND (XTinterface::x11 OR XTinterface::3d)

            ``XT`` terms can also be used to improve the quality of search
            results.  For example, the ``gimp`` package would not usually show
            up when searching the terms ``image editor``.  This can be solved
            using the following technique:

             1. Perform a normal query
             2. Put the first 5 or so results in an Rset
             3. Call Enquire::get_eset using the Rset and an expand filter that
                only accepts ``XT`` terms.  This gives you the tags that are
                most relevant to the query.
             4. Add the resulting terms to the initial query, and search again.

            The Debtags data source is provided by the ``debtags`` package.
            """
        )

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        for tag in self.db.tags_of_package(pkg.name):
            document.add_term("XT"+tag)

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        for tag in self.db.tags_of_package(pkg["Package"]):
            document.add_term("XT"+tag)

def init(**kw):
    """
    Create and return the plugin object.
    """
    if not os.path.exists(DEBTAGSDB):
        return None
    return Debtags()



#!/usr/bin/env python3

import os.path
import sys
from distutils.core import setup

for line in open(os.path.join(os.path.dirname(sys.argv[0]),'update-apt-xapian-index')):
    if line.startswith('VERSION='):
        version = eval(line.split('=')[-1])

setup(name='apt-xapian-index',
      version=version,
      description='Xapian index of Debian packages',
#      long_description=''
      author='Enrico Zini',
      author_email='enrico@debian.org',
      url='http://www.enricozini.org/sw/apt-xapian-index/',
      ## install_requires = [
      ##     "debian", "apt", "xapian",
      ## ],
      license='GPL',
      platforms='any',
      packages=['axi'],
#     py_modules=[''],
      scripts=['update-apt-xapian-index', 'axi-cache'],
     )



try:
    import apt
    import apt_pkg
    HAS_APT=True
except ImportError:
    HAS_APT=False
import os
import os.path
import re

import xapian


class Descriptions:
    def info(self):
        """
        Return general information about the plugin.

        The information returned is a dict with various keywords:

         timestamp (required)
           the last modified timestamp of this data source.  This will be used
           to see if we need to update the database or not.  A timestamp of 0
           means that this data source is either missing or always up to date.
         values (optional)
           an array of dicts { name: name, desc: description }, one for every
           numeric value indexed by this data source.

        Note that this method can be called before init.  The idea is that, if
        the timestamp shows that this plugin is currently not needed, then the
        long initialisation can just be skipped.
        """
        res = dict(
                timestamp=0,
                prefixes=[
                    dict(idx="Z", qp=None, type=None,
                         desc="Stemmed forms of keywords",
                         ldesc="This contains the stemmed forms of keywords as generated by"
                               " TermGenerator and matched by QueryParser"),
                ],
        )
        if not HAS_APT: return res
        if not hasattr(apt_pkg, "config"): return res
        fname = apt_pkg.config.find_file("Dir::Cache::pkgcache")
        if not os.path.exists(fname): return res
        res["sources"] = [dict(path=fname, desc="APT index")]
        res["timestamp"] = os.path.getmtime(fname)
        return res

    def init(self, info, progress):
        """
        If needed, perform long initialisation tasks here.

        info is a dictionary with useful information.  Currently it contains
        the following values:

          "values": a dict mapping index mnemonics to index numbers

        The progress indicator can be used to report progress.
        """
        self.stemmer = xapian.Stem("english")
        self.indexer = xapian.TermGenerator()
        self.indexer.set_stemmer(self.stemmer)

    def send_extra_info(self, db=None, **kw):
        """
        Receive extra parameters from the indexer.

        This may be called more than once, but after init().

        We are using this to get the database instance
        """
        if db is not None:
            self.indexer.set_flags(xapian.TermGenerator.FLAG_SPELLING)
            self.indexer.set_database(db)

    def doc(self):
        """
        Return documentation information for this data source.

        The documentation information is a dictionary with these keys:
          name: the name for this data source
          shortDesc: a short description
          fullDoc: the full description as a chapter in ReST format
        """
        return dict(
            name = "Package descriptions",
            shortDesc = "terms extracted from the package descriptions using Xapian's TermGenerator",
            fullDoc = """
            The Descriptions data source simply uses Xapian's TermGenerator to
            tokenise and index the package descriptions.

            Currently this creates normal terms as well as stemmed terms
            prefixed with ``Z``.
            """
        )

    def index(self, document, pkg):
        """
        Update the document with the information from this data source.

        document  is the document to update
        pkg       is the python-apt Package object for this package
        """
        self.indexer.set_document(document)

        # Index the record
        self.indexer.index_text_without_positions(pkg.name)
        version = pkg.candidate
        if version is not None:
            self.indexer.index_text_without_positions(version.raw_description)
    if not HAS_APT:
        def index(self, document, pkg):
            pass

    def indexDeb822(self, document, pkg):
        """
        Update the document with the information from this data source.

        This is alternative to index, and it is used when indexing with package
        data taken from a custom Packages file.

        document  is the document to update
        pkg       is the Deb822 object for this package
        """
        self.indexer.set_document(document)

        # Index the record
        self.indexer.index_text_without_positions(pkg["Package"])

        if 'Description' in pkg:
            self.indexer.index_text_without_positions(pkg["Description"])
        else:
            # check if we have a translated description
            for k in pkg:
                if k.startswith('Description-'):
                    self.indexer.index_text_without_positions(pkg[k])
                    break

def init(**kw):
    """
    Create and return the plugin object.
    """
    return Descriptions()
